package com.android.launcher3.dragndrop

import android.graphics.Bitmap
import android.graphics.Canvas
import android.graphics.Color
import android.graphics.Matrix
import android.graphics.Paint
import android.graphics.RectF
import android.graphics.Shader
import android.view.MotionEvent
import android.view.VelocityTracker
import com.android.launcher3.DragView
import com.android.launcher3.Launcher
import com.teslacoilsw.kotlin.resetAfter
import com.teslacoilsw.launcher.icon.ShaderMaker
import com.teslacoilsw.launcher.icon.renderAdaptiveShadow

class AdaptiveIconDragView(launcher: Launcher, private val bgShader: ShaderMaker, private val fgShader: ShaderMaker, private val overlayShader: ShaderMaker?, private val bgAnimationMode: AdaptiveBgAnimationMode, bmp: Bitmap, registrationX: Int, registrationY: Int, initialScale: Float, initialMask: Bitmap? = null) : DragView(launcher, bmp, registrationX, registrationY, initialScale) {

    private val maskPaint = Paint(Paint.ANTI_ALIAS_FLAG or Paint.FILTER_BITMAP_FLAG or Paint.DITHER_FLAG).apply {
        color = Color.WHITE
    }
    private val bgPaint = Paint(Paint.ANTI_ALIAS_FLAG or Paint.FILTER_BITMAP_FLAG or Paint.DITHER_FLAG).apply {
        shader = bgShader?.getShader()
    }
    private val fgPaint = Paint(Paint.ANTI_ALIAS_FLAG or Paint.FILTER_BITMAP_FLAG or Paint.DITHER_FLAG).apply {
        shader = fgShader?.getShader()
    }
    private val overlayPaint = Paint(Paint.ANTI_ALIAS_FLAG or Paint.FILTER_BITMAP_FLAG or Paint.DITHER_FLAG).apply {
        shader = overlayShader?.getShader()
    }
    private val tmpMatrix = Matrix()
    private val mask: Bitmap
    private val drawShadow: Boolean

    init {
        if (initialMask != null) {
            mask = initialMask
            drawShadow = false
        } else {
            drawShadow = true
            mask = Bitmap.createBitmap(mBitmap.width, mBitmap.height, Bitmap.Config.ALPHA_8).apply {
                val c = Canvas(this)
                c.drawBitmap(mBitmap, 0f, 0f, maskPaint)
                c.setBitmap(null)
            }
        }

        if (bgShader.bgColor() != Color.TRANSPARENT)
            maskPaint.color = bgShader.bgColor()
    }


    private fun velocityToDisplacement(velocity: Float): Float {
        val clampedVelocity = velocity.coerceIn(-1000f, 1000f)
        return Math.min(mBitmap.width, mBitmap.height) * clampedVelocity / -1000f
    }

    private val velocityTracker = VelocityTracker.obtain()
    private var displacementX = 0f
    private var displacementY = 0f
    override fun move(touchX: Int, touchY: Int, ev: MotionEvent?) {
        super.move(touchX, touchY, ev)
        if (ev == null)
            velocityTracker.clear()
        else
            velocityTracker.addMovement(ev)
        velocityTracker.computeCurrentVelocity(1000)
        displacementX = velocityToDisplacement(velocityTracker.xVelocity)
        displacementY = velocityToDisplacement(velocityTracker.yVelocity)
        invalidate()
    }

    private fun applyMatrix(shader: Shader, shaderIntrinsicSize: Float, shaderViewportSize: Float, dx: Float, dy: Float, dstRect: RectF? = null) {
        if (shaderIntrinsicSize == 0f)
            return
        val size = mBitmap.width

        val scale = size / shaderIntrinsicSize
        val offset = 0f


        tmpMatrix.reset()
        tmpMatrix.setScale(scale, scale)
        if (shaderIntrinsicSize != shaderViewportSize) {
            val scale = shaderIntrinsicSize / shaderViewportSize
            tmpMatrix.postScale(scale, scale, size / 2f, size / 2f)
        }
        tmpMatrix.postTranslate(dx-offset, dy-offset)
        shader.setLocalMatrix(tmpMatrix)
        if (dstRect != null)
            tmpMatrix.mapRect(dstRect)
    }

    private val ADAPTIVE_BASE_SIZE = 72
    private val ADAPTIVE_LAYER_SIZE = 108
    private val ADAPTIVE_LAYER_TO_ICON_SIZE_RATIO = ADAPTIVE_LAYER_SIZE / ADAPTIVE_BASE_SIZE.toFloat()
    private val ADAPTIVE_OFFSET_RATIO = (ADAPTIVE_LAYER_SIZE - ADAPTIVE_BASE_SIZE)/2f/ADAPTIVE_LAYER_SIZE;
    private val BG_DISPLACEMENT_FACTOR = .08f
    private val FG_DISPLACEMENT_FACTOR = .12f
    override fun onDraw(canvas: Canvas) {
        mHasDrawn = true

        if (drawShadow)
            renderAdaptiveShadow(canvas)


        if (bgAnimationMode == AdaptiveBgAnimationMode.NONE || bgShader.bgColor() != Color.TRANSPARENT) {
            canvas.drawBitmap(mask, 0f, 0f, maskPaint)
        }

        val dx = displacementX * FG_DISPLACEMENT_FACTOR
        val dy = displacementY * FG_DISPLACEMENT_FACTOR

        if (bgPaint.shader != null) {
            val bgDx: Float
            val bgDy: Float
            when (bgAnimationMode) {
                AdaptiveBgAnimationMode.NONE -> {
                    bgDx = 0f; bgDy = 0f;
                } AdaptiveBgAnimationMode.WITH_FG -> {
                    bgDx = dx; bgDy = dy
                } AdaptiveBgAnimationMode.NORMAL -> {
                    bgDx = displacementX * BG_DISPLACEMENT_FACTOR
                    bgDy = displacementY * BG_DISPLACEMENT_FACTOR
                }
            }
            applyMatrix(bgPaint.shader, bgShader.getIntrinsicSize(), bgShader.getViewportSize(), bgDx, bgDy)
//            canvas.drawRect(0f, 0f, measuredWidth.toFloat(), measuredHeight.toFloat(), bgPaint)
            canvas.drawBitmap(mask, 0f, 0f, bgPaint)
        }

        if (fgShader != null) {
            val rectF = RectF(0f, 0f, fgShader.getIntrinsicSize(), fgShader.getIntrinsicSize())
            applyMatrix(fgPaint.shader, fgShader.getIntrinsicSize(), fgShader.getViewportSize(), dx, dy, rectF)
            canvas.resetAfter {
                canvas.clipRect(rectF)
                canvas.drawBitmap(mask, 0f, 0f, fgPaint)
            }
        }

        if (overlayShader != null) {
            applyMatrix(overlayPaint.shader, overlayShader.getIntrinsicSize(), overlayShader.getViewportSize(), 0f, 0f)
            canvas.resetAfter {
                canvas.drawBitmap(mask, 0f, 0f, overlayPaint)
            }
        }

        if (mTintColor != Color.TRANSPARENT) {
            val oldColor = maskPaint.color
            maskPaint.color = mTintColor
            canvas.drawBitmap(mask, 0f, 0f, maskPaint)
            maskPaint.color = oldColor
        }

    }
}

enum class AdaptiveBgAnimationMode {
    NONE,
    NORMAL,
    WITH_FG,
}
