package com.teslacoilsw.launcher.deepshortcutproxy

import android.annotation.TargetApi
import android.content.ComponentName
import android.content.ContentProvider
import android.content.ContentResolver
import android.content.ContentValues
import android.content.Context
import android.content.pm.LauncherApps
import android.database.Cursor
import android.database.MatrixCursor
import android.graphics.Bitmap
import android.net.Uri
import android.os.Build
import android.os.ParcelFileDescriptor
import com.android.launcher3.Utilities
import com.android.launcher3.compat.UserHandleCompat
import com.android.launcher3.compat.UserManagerCompat
import com.android.launcher3.config.ProviderConfig
import com.teslacoilsw.launcher.BuildConfig
import com.teslacoilsw.launcher.util.PackageNames
import timber.log.Timber
import java.io.File
import java.io.FileNotFoundException
import java.io.FileOutputStream


private val sanitizeRegex = "[^a-zA-Z0-9\\._]+".toRegex()
private fun sanitize(s: String) : String {
    if (s.isBlank() || s == "..")
        return "_"
    return s.replace(sanitizeRegex, "_")
}

@TargetApi(Build.VERSION_CODES.N_MR1)
class DSPProvider : ContentProvider() {

    val iconCachePath by lazy { File(context.cacheDir, "deepShortcutIcons") }

    override fun delete(uri: Uri, selection: String?, selectionArgs: Array<String>?): Int {
        // Implement this to handle requests to delete one or more rows.
        throw UnsupportedOperationException("Not yet implemented")
    }

    override fun getType(uri: Uri): String? {
        // TODO: Implement this to handle requests for the MIME type of the data
        // at the given URI.
        throw UnsupportedOperationException("Not yet implemented")
    }

    override fun insert(uri: Uri, values: ContentValues?): Uri? {
        throw UnsupportedOperationException("Not yet implemented")
    }

    override fun onCreate(): Boolean {
        return true
    }

    private fun isAuthorized(): Boolean {
        return PackageNames.SESAME == callingPackage || BuildConfig.APPLICATION_ID == callingPackage;
    }
    override fun query(uri: Uri, projection: Array<String>?, selection: String?,
                       selectionArgs: Array<String>?, sortOrder: String?): Cursor? {
        if (!isAuthorized()) {
            Timber.tag("Nova.DSP").e("Rejecting unauthorized package `$callingPackage` from accessing DeepShortcutProxyProvider")
            return null
        }

        val apps = context.getSystemService(Context.LAUNCHER_APPS_SERVICE) as LauncherApps
        val query = LauncherApps.ShortcutQuery()
        query.setQueryFlags(LauncherApps.ShortcutQuery.FLAG_MATCH_DYNAMIC or LauncherApps.ShortcutQuery.FLAG_MATCH_MANIFEST)

        val list = apps.getShortcuts(query, UserHandleCompat.myUserHandle().user) ?: emptyList()

        val userManager = UserManagerCompat.getInstance(context)
        val c = MatrixCursor(arrayOf("package", "id", "label", "longLabel", "activity", "categories", "disabledMessage", "extras", "lastChangedTimestamp", "rank", "userHandleId", "hasKeyFieldsOnly", "isDeclaredInManifest", "isDynamic", "isEnabled", "isImmutable", "isPinned", "iconUri"))
        for (i in list.indices) {
            val s = list[i]
            val user = UserHandleCompat.fromUser(s.userHandle)
            val resIconId = if (s.reflectHasIconResource()) s.reflectGetIconResourceId() else 0
            val iconUri = when {
                resIconId != 0 -> Uri.Builder().scheme(ContentResolver.SCHEME_ANDROID_RESOURCE).authority(s.`package`).appendPath(resIconId.toString())
                else -> Uri.Builder().scheme(ContentResolver.SCHEME_CONTENT).authority(ProviderConfig.DEEPSHORTCUT_PROXY_AUTHORITY).appendPath("icon").appendPath(userManager.getSerialNumberForUser(user).toString()).appendPath(s.`package`).appendPath(Uri.encode(s.id))
            }
            c.addRow(arrayOf(s.`package`, s.id, s.shortLabel, s.longLabel, flattenCn(s.activity), "TODO", s.disabledMessage, s.extras, s.lastChangedTimestamp, s.rank, userManager.getSerialNumberForUser(user), s.hasKeyFieldsOnly(), s.isDeclaredInManifest, s.isDynamic, s.isEnabled, s.isImmutable, s.isPinned, iconUri))
        }
        return c
    }

    private fun flattenCn(cn: ComponentName?): String? {
        return cn?.flattenToShortString()
    }

    override fun update(uri: Uri, values: ContentValues?, selection: String?,
                        selectionArgs: Array<String>?): Int {
        // TODO: Implement this to handle requests to update one or more rows.
        throw UnsupportedOperationException("Not yet implemented")
    }

    override fun openFile(uri: Uri, mode: String): ParcelFileDescriptor? {
        if (!isAuthorized()) {
            Timber.tag("Nova.DSP").e("Rejecting unauthorized package `$callingPackage` from accessing DeepShortcutProxyProvider")
            return null
        }
        val path = uri.pathSegments
        if (path.size != 4 || path[0] != "icon")
            throw FileNotFoundException("Invalid path $uri, should start with /icon/")

        val userId = path[1].toLong()
        val pkg = path[2]

        var id = Uri.decode(path[3])
        if (id.isBlank() || id == "." || id == "..")
            id = "_"

        val localPath = "$userId/${Uri.encode(pkg)}/${Uri.encode(id)}"
        if (localPath.contains("/../"))
            throw FileNotFoundException("Invalid path $uri, should not contain '/../'")

        val cachedIcon = File(iconCachePath, localPath)
        if (cachedIcon.exists())
            return ParcelFileDescriptor.open(cachedIcon, ParcelFileDescriptor.MODE_READ_ONLY)

        val apps = context.getSystemService(Context.LAUNCHER_APPS_SERVICE) as LauncherApps

        val q = LauncherApps.ShortcutQuery()
        q.setPackage(pkg)
        q.setShortcutIds(listOf(id))
        q.setQueryFlags(LauncherApps.ShortcutQuery.FLAG_MATCH_DYNAMIC or LauncherApps.ShortcutQuery.FLAG_MATCH_MANIFEST)
        val user = UserManagerCompat.getInstance(context).getUserForSerialNumber(userId).user
        val shortcuts = apps.getShortcuts(q, user) ?: emptyList()
        if (shortcuts.size > 0) {
            val shortcut = shortcuts.first()
            if (shortcut.reflectHasIconFile()) {
                val fileFd = apps.reflectionGetShortcutIconFd(shortcut)
                if (fileFd != null) {
                    return fileFd
                }
            }

            val d = apps.getShortcutIconDrawable(shortcut, context.resources.displayMetrics.densityDpi)
            val bmp = Utilities.createIconBitmap(d, context)
            cachedIcon.parentFile.mkdirs()
            bmp.compress(Bitmap.CompressFormat.PNG, 100, FileOutputStream(cachedIcon))
            return ParcelFileDescriptor.open(cachedIcon, ParcelFileDescriptor.MODE_READ_ONLY)
        }
        throw FileNotFoundException("No icon exists at " + uri)
    }
}
